use std::path::PathBuf;
use std::{fs::create_dir_all, io::Write, path::Path};
use std::fs::{remove_file, DirEntry, OpenOptions};

use directories::ProjectDirs;
use iced::futures::join;
use reqwest::Client;
use reqwest::header::{HeaderMap, HeaderValue, ACCEPT_ENCODING};
use serde::{Deserialize, Serialize};

const KURO_BASE_URL: &str = "https://prod-alicdn-gamestarter.kurogame.com/launcher/launcher/50004_obOHXFrFanqsaIEOmuKroCcbZkQRBC7c/G153";

#[derive(Serialize, Deserialize, Debug)]
struct BackgroundFunctionCode {
    background: String
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all(serialize = "camelCase", deserialize = "camelCase"))]
struct KuroVersionIndex {
    function_code: BackgroundFunctionCode
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all(serialize = "camelCase", deserialize = "camelCase"))]
struct KuroBackgroundInformation {
    function_switch: i32,
    background_file: String,
    background_file_type: i32,
    first_frame_image: String,
    slogan: String
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all(serialize = "camelCase", deserialize = "camelCase"))]
struct HoyoImageInformation {
    url: String,
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all(serialize = "camelCase", deserialize = "camelCase"))]
struct HoyoDisplayInformation {
    name: String,
    icon: HoyoImageInformation,
    background: HoyoImageInformation
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all(serialize = "camelCase", deserialize = "camelCase"))]
struct HoyoGameInformation {
    display: HoyoDisplayInformation
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all(serialize = "camelCase", deserialize = "camelCase"))]
struct HoyoGameList {
    games: Vec<HoyoGameInformation>
}

#[derive(Serialize, Deserialize, Debug)]
#[serde(rename_all(serialize = "camelCase", deserialize = "camelCase"))]
struct HoyoLauncherInformation {
    data: HoyoGameList
}

pub async fn refresh_install() -> Result<(), String> {
    let reqwest_client = build_client()?;
    let proj_dirs = ProjectDirs::from("com", "RabbyDevs", "rr-launcher")
        .ok_or_else(|| "Failed to get project directories".to_string())?;
    
    let (kuro_result, hoyo_result) = join!(
        refresh_kuro_install(&proj_dirs, &reqwest_client),
        refresh_hoyo_install(&proj_dirs, &reqwest_client)
    );
    
    if let Err(e) = kuro_result {
        eprintln!("Error in Kuro install: {}", e);
    }
    
    if let Err(e) = hoyo_result {
        eprintln!("Error in Hoyo install: {}", e);
    }
    
    Ok(())
}

fn build_client() -> Result<Client, String> {
    let mut headers = HeaderMap::new();
    headers.insert(ACCEPT_ENCODING, HeaderValue::from_static("gzip, deflate, br"));

    Client::builder()
        .default_headers(headers)
        .gzip(true) 
        .build()
        .map_err(|e| format!("Failed to build HTTP client: {}", e))
}

async fn refresh_kuro_install(proj_dirs: &ProjectDirs, client: &Client) -> Result<(), String> {
    let data_dir_buf = proj_dirs.data_dir().join("kuro/wuwa");
    let data_dir = data_dir_buf.as_path();
    
    create_dir_all(data_dir)
        .map_err(|e| format!("Failed to create directory: {}", e))?;
    
    let kuro_url = KURO_BASE_URL.to_string();
    let index_url = format!("{}{}", kuro_url, "/index.json");
    
    eprintln!("Fetching index from: {}", index_url);
    
    let response = client
        .get(&index_url)
        .send()
        .await
        .map_err(|e| format!("Failed to send HTTP request: {}", e))?;
    
    if !response.status().is_success() {
        return Err(format!("HTTP request failed with status {}", response.status()));
    }
    
    eprintln!("Response headers: {:?}", response.headers());

    let version_index: KuroVersionIndex = response
        .json()
        .await
        .map_err(|e| {
            format!("Failed to parse JSON - {}", e)
        })?;
    
    eprintln!("Successfully parsed index.json: {:?}", version_index);

    let bg_info_url = format!("{}{}{}{}", 
        "https://prod-alicdn-gamestarter.kurogame.com/launcher/50004_obOHXFrFanqsaIEOmuKroCcbZkQRBC7c/G153", 
        "/background/", 
        version_index.function_code.background, 
        "/en.json"
    );
    
    eprintln!("Fetching background info from: {}", bg_info_url);
    
    let bg_response = client
        .get(&bg_info_url)
        .send()
        .await
        .map_err(|e| format!("Failed to send background info request: {}", e))?;
    
    if !bg_response.status().is_success() {
        return Err(format!("Background info request failed with status {}", bg_response.status()));
    }
    
    let bg_information: KuroBackgroundInformation = bg_response
        .json()
        .await
        .map_err(|e| format!("Failed to parse background info JSON: {}", e))?;
    
    eprintln!("Successfully parsed background info: {:?}", bg_information);
    
    update_file_if_needed(data_dir, client, &bg_information.background_file, "background").await?;
    update_file_if_needed(data_dir, client, &bg_information.slogan, "splash").await?;
    update_file_if_needed(data_dir, client, &bg_information.first_frame_image, "first_frame_image").await?;
    
    Ok(())
}

async fn update_file_if_needed(dir: &Path, client: &Client, file_url: &str, file_type: &str) -> Result<(), String> {
	let current_file: Option<DirEntry> = {
		let mut current_file = None;
		for path in dir.read_dir().unwrap() {
			let path = path.unwrap();
			if path.file_name().into_string().unwrap().starts_with(format!("{}_", file_type).as_str()) {
				current_file = Some(path);
			}
		}
		current_file
	};

    let filename = extract_filename_from_url(file_url);
    let expected_file = format!("{}_{}", file_type, filename);
    let file_path = dir.join(&expected_file);

    if let Some(file) = current_file {
        if filename != file.file_name().into_string().unwrap().strip_prefix(format!("{}_", file_type).as_str()).unwrap() {
            update_file(client, file_path, file_type, file_url).await?;
            remove_file(file.path()).unwrap();
        } else {
            eprintln!("{} file already exists at: {}", file_type, file.path().display());
        }
    } else {
        update_file(client, file_path, file_type, file_url).await?;
    }
    
    Ok(())
}

async fn update_file(client: &Client, file_path: PathBuf, file_type: &str, file_url: &str) -> Result<(), String> {
    eprintln!("Downloading {} file from: {}", file_type, file_url);
    let file_bytes = client
        .get(file_url)
        .send()
        .await
        .map_err(|e| format!("Failed to send request for {} file: {}", file_type, e))?
        .bytes()
        .await
        .map_err(|e| format!("Failed to get bytes for {} file: {}", file_type, e))?;
        
    let mut file = OpenOptions::new()
        .write(true)
        .create(true)
        .truncate(true)
        .open(&file_path)
        .map_err(|e| format!("Failed to create {} file ({}): {}", file_type, file_path.display(), e))?;
        
    file.write_all(&file_bytes)
        .map_err(|e| format!("Failed to write {} file: {}", file_type, e))?;
        
    file.flush()
        .map_err(|e| format!("Failed to flush {} file: {}", file_type, e))?;
        
    eprintln!("Successfully downloaded {} file to: {}", file_type, file_path.display());
    
    Ok(())
}

async fn refresh_hoyo_install(proj_dirs: &ProjectDirs, client: &Client) -> Result<(), String> {
    let hoyo_url = "https://sg-hyp-api.hoyoverse.com/hyp/hyp-connect/api/getGames?launcher_id=VYTpXlbWo8&language=en-us";
    
    eprintln!("Fetching Hoyo launcher info from: {}", hoyo_url);
    
    let response = client
        .get(hoyo_url)
        .send()
        .await
        .map_err(|e| format!("Failed to fetch Hoyo launcher info: {}", e))?;
        
    if !response.status().is_success() {
        return Err(format!("Hoyo launcher info request failed with status {}", response.status()));
    }
    
    let hoyo_launcher_info: HoyoLauncherInformation = response
        .json()
        .await
        .map_err(|e| format!("Failed to parse Hoyo launcher info: {}", e))?;
        
    eprintln!("Successfully parsed Hoyo launcher info");

    for game in hoyo_launcher_info.data.games {
        let game_abbreviation = match game.display.name.as_str() {
            "Zenless Zone Zero" => "zzz",
            "Honkai Star Rail" => "hsr",
            "Genshin Impact" => "gi",
            _ => {
                eprintln!("Skipping unknown game: {}", game.display.name);
                continue;
            }
        };
        
        let data_dir_buf = proj_dirs.data_dir().join(format!("hoyoverse/{}", game_abbreviation));
        let data_dir = data_dir_buf.as_path();
        
        create_dir_all(data_dir)
            .map_err(|e| format!("Failed to create directory for {}: {}", game_abbreviation, e))?;
            
        eprintln!("Processing game: {} ({})", game.display.name, game_abbreviation);
        
        let icon_url = &game.display.icon.url;
        update_file_if_needed(data_dir, client, icon_url, "icon").await?;

        let bg_url = &game.display.background.url;
        update_file_if_needed(data_dir, client, bg_url, "background").await?;
    }
    
    Ok(())
}

fn extract_filename_from_url(url: &str) -> String {
    match url.split('/').next_back() {
        Some(filename) => String::from(filename),
        None => String::from("unknown_file"),
    }
}